(** S2DLDraws
@Name: Simple 2D Library - Main Surface Drawing Object
  This file is part of the Free Pascal Simple 2D Libraries
@Version: 1.03
@Note: Do not initialize yourself !!! - Automatically done !!!
@Change_History: ver 1.03 - Added Flip to S2DLDraws to allow calculation of FPS
           - Auto check a config file to see if must run full screen.
@Copyright: (c) 2004-2005 by William Cairns (CairnsGames) and
                         the Delphi Game Development (DGDev) Site

@License: This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
       (I got this from the JEDI-SDL files and it sounds Cool!)
@
By using this work you have immediatly agreed with the above.
@
 The work can be used for any purpose. You may sell it, use it in a game
 delete it or change it as you wish with no reward to the the original
 developers.
@
 We do however ask that if this work is used that some mention of it be made
 on the PGDev web site.

@Usage: Just include this unit as the first S2DL Unit in the Main Program
   S2DLDraws is responsible for setting up the required SDL surface from the Screen
   S2DLDraws will also end the SDL session and clear its own surface

@Author: W.Cairns **)
unit S2DLDraws;

{$ifdef FPC}
{$mode delphi}
{$endif}

interface


Uses
  Classes,
  SysUtils,
  S2DLLogger,
  sdl;

Type
  (**
    TS2DLDraw encapsulated the screen area for the SDL application. A single
    instance of the class is created and used to control the display area.
    Methods and properties are also included to calculate the Frame rate
    currently being seen in the game.
    @
    This class should never be instantiated as it is done automatically in
    the initialization section of the unit.
  **)
  TS2DLDraw = Class(TObject)
  Private
    FSurface : PSDL_Surface;
    Rect : TSDL_Rect;
    LastTick : UInt32;
    LastFPS : Integer;
    CurFPS : Integer;
    FAppPath: String;
    procedure SetAppPath(const Value: String);
  Public
    Property Surface : PSDL_Surface read FSurface;
    Property AppPath : String read FAppPath write SetAppPath;
    Constructor Create;
    Destructor Destroy; Override;
    Procedure Clear;
    Procedure Flip;
    Function FPS : Integer;
    Procedure WindowCaption(NewCaption : String);
  End;

Var
  (**S2DLDraw is instantiated by the S2DL code and can be used in the
  program.**)
  S2DLDraw : TS2DLDraw;
  Error : Boolean;
  
Function MakeSDLColor(R,G,B : UInt8) : TSDL_Color;

implementation

(**
  Create an TSDL_Color from the basic colors

  @param   R as an UInt8
  @param   G as an UInt8
  @param   B as an UInt8
  @return  a TSDL_Color
**)
Function MakeSDLColor(R,G,B : UInt8) : TSDL_Color;
Begin
  Result.R := R;
  Result.G := G;
  Result.B := B;
End;

{ TS2DLDraw }

(**

  FullScreen checks the Application Configuration file (automatically) to see
  if the program must run as a full screen application.

  @return  a Boolean

**)
Function FullScreen : Boolean;
Var
  R : Boolean;
  RS : String;
  FApplicationPath : String;
  FApplicationName : String;
  N : String;
  L : TStringList;
Begin
  R := False;
  FApplicationName := ExtractFileName( ParamStr(0) );
  FApplicationPath := ExtractFilePath( ParamStr(0) );
  N := FApplicationPath + Copy( FApplicationName, 0, Length( FApplicationName ){$IFDEF WIN32} - 4{$ENDIF} ) + '.cfg';
  If FileExists(N) then
    begin
      L := TStringList.Create;
      L.LoadFromFile(FApplicationPath + Copy( FApplicationName, 0, Length( FApplicationName ){$IFDEF WIN32} - 4{$ENDIF} ) + '.cfg');
      RS := L.Values['FullScreen'];
      If RS = '' then
        RS := 'N';
      If RS[1] in ['Y','y'] then
        R := True;
      L.Free;
    End;
  Result := R;
End;

(**
  This is the constructor method for the TS2DLDraw class.
**)
Constructor TS2DLDraw.Create;
Var
  Flags : Integer;
Begin
  LastFPS := 0;
  CurFPS := 0;
  Error := false;
  if SDL_Init(SDL_INIT_VIDEO) = -1 then
  begin
    Log.LogError(Format('Could not initialize SDL : %s',
      [SDL_GetError]), 'S2DLDraw.Create');
    SDL_Quit;
    halt(1);
  end;

  // To set up full screen mode add
  //     or SDL_FULLSCREEN
  If FullScreen then
    Flags := SDL_SWSURFACE or SDL_HWPALETTE or SDL_FULLSCREEN
  Else
    Flags := SDL_SWSURFACE or SDL_HWPALETTE;

  FSurface := SDL_SetVideoMode(640, 480, 16, Flags);
  if FSurface = nil then
  begin
    Log.LogError(Format('Couldn''t set 640x480x16 video mode : %s',[SDL_GetError]), 'S2DLDraw.Create');
    Error := true;
  end;

  Rect.X := 0;
  Rect.Y := 0;
  Rect.W := FSurface.W;
  Rect.H := FSurface.H;
  LastTick := SDL_GetTicks;
  FAppPath := ExtractFilePath( ParamStr(0) );
End;


(**
  This is the destructor method for the TS2DLDraw class.
**)
Destructor TS2DLDraw.Destroy;
Begin
  Log.LogStatus('S2DLDraw surface being cleared','TS2DLDraw.Destroy');
  SDL_FreeSurface(FSurface);
End;

(**
  Clear the SDL Surface
**)
procedure TS2DLDraw.Clear;
begin
  SDL_FillRect(FSurface, @Rect, 0);
end;

(**
  Flips the SDL surface and does the necessary calculations to calculate the
  current FPS.
**)
procedure TS2DLDraw.Flip;
begin
  CurFPS := CurFPS + 1;
  If SDL_GetTicks > LastTick + 1000 then
    Begin
      LastFPS := CurFPS;
      CurFPS := 0;
      LastTick := SDL_GetTicks;
    End;
  SDL_Flip(S2DLDRaw.Surface);
end;

(**Returns the current Frame Per Second calculation. 

  @return  an Integer
**)
function TS2DLDraw.FPS: Integer;
begin
  If LastFPS = 0 then
    Begin
      If CurFPS = 0 then
        Result := 1
      Else
        Result := CurFPS
    End
  Else
    Result := LastFPS;
end;

procedure TS2DLDraw.WindowCaption(NewCaption: String);
begin
  SDL_WM_SetCaption(PChar(NewCaption), nil);
end;

procedure TS2DLDraw.SetAppPath(const Value: String);
begin
  FAppPath := Value;
end;

Initialization
  // This initalizes the SDL session and creates the SDL surface for Drawing
  Log.LogStatus('SDL being started up.','S2DLDraws.Initialization');
  S2DLDraw := TS2DLDraw.Create;

Finalization
  // This shut downs the SDL session and will fre the SDL surface
  Log.LogStatus('SDL being closed down.','S2DLDraws.Finalization');
  S2DLDraw.Free;
  SDL_Quit;

end.
