(** S2DLImage
@Name: Simple 2D Library - Images Unit
  This file is part of the Free Pascal Simple 2D Libraries
@Version: 1.04
@Change_History: ver 1.04 - Added integration with Windows GDI functions through Bitmap
@ - Modified to Load BMP, JPG, GIF etc

@Copyright: (c) 2004-2005 by William Cairns (CairnsGames) and
                         the Delphi Game Development (DGDev) Site

@License: This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
       (I got this from the JEDI-SDL files and it sounds Cool!)
@
By using this work you have immediatly agreed with the above.
@
 The work can be used for any purpose. You may sell it, use it in a game
 delete it or change it as you wish with no reward to the the original
 developers.
@
 We do however ask that if this work is used that some mention of it be made
 on the PGD web site.

@Usage: A single S2DLImageList is created by default at the start of the
  program. This imagelist can be accessed by Images.<method>  @
  Additional S2DLImageLists can be created if required.

@Author: W.Cairns **)
Unit S2DLImage;

{$ifdef FPC}
{$mode delphi}
{$endif}

Interface
{ TODO : Move Copy Image into seperate unit (A utils unit or something) }               

Uses
  Classes,
  Graphics,
  SysUtils,
  S2DLLogger,
  sdl,
  SDL_Image,
  S2DLDraws;

Type
  (** Forward **)
  TS2DLImage = Class;
  (**
    TS2DLImageList is an image repository where images can be stored and accessed
    as needed in the program. Each image stored in the list is stored as a
    TS2DLImage. Includes functionality to @ - Search @ - Delete @ - Load Images 
  **)
  TS2DLImageList = Class
    Private
      FItems : TList;
      Function GetImage(Index : Integer) : TS2DLImage;
      Function GetImageByName(Name : String) : TS2DLImage;
    Public
      Property Items[Index : Integer] : TS2DLImage read GetImage; default;
      Property ItemByName[Name : String] : TS2DLImage read GetImageByName;
      Constructor Create;
      Destructor Destroy; Override;
      Function Find(Name : String) : TS2DLImage;
      Function FindID(Name : String) : Integer;
      Procedure Add(Item : TS2DLImage);
      Function Count : Integer;
      Function Delete(Item : Integer) : Boolean;
      Procedure LoadImage(Name : String; FileName : String); Overload;
      Procedure LoadImage(Name : String; FileName : String; R,G,B : Integer); Overload;
  End;
  (**
    TS2DLImage encapsulates the SDL surface functionality into a class that
    can more easily be used in a program. TS2DLImage is used as an image
    respository and called to draw the image onto the screen.
  **)
  TS2DLImage = Class
  private
    function GetHeight: Integer;
    function GetWidth: Integer;
  public
    Name : String;
    ImageSurface: PSDL_Surface;
    SrcRect, DestRect: TSDL_Rect;
    PatternWidth, PatternHeight : Integer;
    Property Width : Integer read GetWidth;
    Property Height : Integer read GetHeight;
    Constructor Create; 
    Constructor CreateSizedSurface(Width,Height,ColorDepth : Integer);
    Destructor Destroy; Override;
    Procedure LoadFromFile(FileName : PChar);
    Procedure Draw(S2DLDraw : TS2DLDraw; X,Y : Integer); Overload;
    Procedure Draw(S2DLDraw : TS2DLDraw; X,Y : Integer; Pattern : Integer); Overload;
    Procedure Draw(X,Y : Integer); Overload;
    Procedure Draw(X,Y : Integer; Pattern : Integer); Overload;
    Procedure Draw(Surface : PSDL_Surface; X,Y : Integer); Overload;
    Procedure Draw(Surface : PSDL_Surface; X,Y : Integer; Pattern : Integer); Overload;
    Procedure DrawTiled(S2DLDraw : TS2DLDraw; X,Y,W,H : Integer); Overload;
    Procedure DrawTiled(S2DLDraw : TS2DLDraw; X,Y,W,H,Pattern : Integer); Overload;
    Procedure DrawTiled(X,Y,W,H : Integer); Overload;
    Procedure DrawTiled(X,Y,W,H,Pattern : Integer); Overload;
    Procedure DrawTiled(Surface : PSDL_Surface; X,Y,W,H : Integer); Overload;
    Procedure DrawTiled(Surface : PSDL_Surface; X,Y,W,H,Pattern : Integer); Overload;
    Procedure TransparentColor(R,G,B : Byte);
    Procedure SetPixel(X,Y : Integer; Color : UInt32);
    Procedure CopyBitmap(X,Y : Integer; BMP : TBitmap; TransparentColor : UInt32);
  End;

Var
  Images : TS2DLImageList;

Implementation

{ TS2DLImageList }

(**
  This is the constructor method for the TS2DLImageList class. Creates the list to
  store the images in.
**)
Constructor TS2DLImageList.Create;
Begin
  FItems := TList.Create;
End;

Destructor TS2DLImageList.Destroy;
Var
  I : Integer;
Begin
  // Free memory used for each image
  For I := FItems.Count - 1 downto 0 do
    Begin
      TS2DLImage(FItems[I]).Free;
      FItems.Delete(I);
    End;
  FItems.Free;
End;

Function TS2DLImageList.GetImage(Index : Integer) : TS2DLImage;
Begin
  Result := TS2DLImage(FItems[Index]);
End;

Function TS2DLImageList.GetImageByName(Name : String) : TS2DLImage;
Var
  I : Integer;
Begin
  For I := FItems.Count-1 downto 0 do
    Begin
      If TS2DLImage(FItems[I]).Name = Name then
        Begin
          Result := TS2DLImage(FItems[I]);
          Exit;
        End;
    End;
  Result := Nil;
End;

(**
  Find an image based on its Name.

  @param   Name as a String
  @return  a TS2DLImage
**)
Function TS2DLImageList.Find(Name : String) : TS2DLImage;
Begin
  Result := GetImageByName(Name);
End;

(**
  Adds a new image into the Images List.

  @param   Item as a TS2DLImage
**)
Procedure TS2DLImageList.Add(Item : TS2DLImage);
Begin
  FItems.Add(Item);
End;

(**
  Returns number of images stored in the image list.

  @return  an Integer
**)
Function TS2DLImageList.Count : Integer;
Begin
  Result := FItems.Count;
End;

(**
  Removes and frees an image from the list.

  @param   Item as an Integer
  @return  a Boolean
**)
Function TS2DLImageList.Delete(Item : Integer) : Boolean;
Var
  GameImage : TS2DLImage;
Begin
  GameImage := TS2DLImage(FItems[Item]);
  FItems.Delete(Item);
  GameImage.Free;
  Result := True;
End;

procedure TS2DLImageList.LoadImage(Name: String; FileName: String);
Var
  Image : TS2DLImage;
Begin
  // Start Loading Images
  Log.LogStatus('Loading:'+Name,'ImageList.LoadImage');
  Image := TS2DLImage.Create;
  Image.LoadFromFile(PChar(FileName));
  Image.TransparentColor(0,0,0);
  Image.Name := Name;
  Add(Image);
end;


procedure TS2DLImageList.LoadImage(Name, FileName: String; R, G,
  B: Integer);
Var
  Image : TS2DLImage;
begin
  LoadImage(Name,FileName);
  Image := GetImageByName(Name);
  Image.TransparentColor(R,G,B);
end;

(**
  Find an Image based on its name. But returns the index number of the image in the list.

  @param   Name as a String
  @return  an Integer
**)
function TS2DLImageList.FindID(Name: String): Integer;
Var
  I : Integer;
Begin
  For I := FItems.Count-1 downto 0 do
    Begin
      If TS2DLImage(FItems[I]).Name = Name then
        Begin
          Result := I;
          Exit;
        End;
    End;
  Result := -1;
end;

{ TS2DLImage }

(**
  This is a getter method for the Height property. Returns effective height
  used by the image when drawing - if Patetrns are set returns PatternHeight,
  else returns Surface Height.

  @return  an Integer
**)
function TS2DLImage.GetHeight: Integer;
begin
  If PatternHeight > 0 then
    Result := PatternHeight
  Else
    Result := ImageSurface.H;
end;

(**

  This is a getter method for the Width property. Returns effective Width
  used by the image when drawing - if Patetrns are set returns PatternWidth,
  else returns Surface Width.

  @return  an Integer

**)
function TS2DLImage.GetWidth: Integer;
begin
  If PatternWidth > 0 then
    Result := PatternWidth
  Else
    Result := ImageSurface.W;
end;

(**

  This is the constructor method for the TS2DLImage class.

**)
Constructor TS2DLImage.Create;
Begin
End;

(**
  This is the constructor method for the TS2DLImage class. It creates a surface with
  size and color depth defined. This is useful for creating a temporary image
  that a background or other composite image can be created on.

  @param   Width      as an Integer
  @param   Height     as an Integer
  @param   ColorDepth as an Integer
**)
Constructor TS2DLImage.CreateSizedSurface(Width,Height,ColorDepth : Integer);
Begin
  ImageSurface := SDL_CreateRGBSurface(SDL_SWSURFACE or SDL_HWPALETTE, Width,Height,
      ColorDepth, 0, 0, 0, 0);
End;

(**
  This is the destructor method for the TS2DLImage class. Frees the memory
  allocated to the image surface.
**)
Destructor TS2DLImage.Destroy;
Begin
  SDL_FreeSurface(ImageSurface);
End;

Procedure TS2DLImage.LoadFromFile(FileName : PChar);
Begin
  If Name = '' then
    Name := FileName;
  Log.LogStatus('Loading Image: '+FileName, 'TS2DLImage.LoadFromFile');
//  ImageSurface := SDL_LoadBMP(FileName);
  ImageSurface := IMG_LOAD(FileName);
  if ImageSurface = nil then
    begin
//      Log.LogError(Format('Couldn''t load image : %s',[SDL_GetError]), 'TS2DLImage.LoadFromFile');
      Exit;
    end;
  SRCRect.X := 0; SrcRect.Y := 0;
  SrcRect.W := ImageSurface.W;
  SrcRect.H := ImageSurface.H;
End;

(**
  Draws the Surface onto the Screen. Has been overridden to support multiple
  Different parameter collections. @ This implementation of draw enables the
  display of the whole image at position X,Y of the S2DLDraw surface.

  @param   S2DLDraw as a TS2DLDraw
  @param   X        as an Integer
  @param   Y        as an Integer

**)
Procedure TS2DLImage.Draw(S2DLDraw : TS2DLDraw; X,Y : Integer);
Begin
  DestRect.x := X;
  DestRect.y := Y;
  SrcRect.X := 0;
  SrcRect.Y := 0;
  SrcRect.W := ImageSurface.W;
  SrcRect.H := ImageSurface.H;
  SDL_BlitSurface(ImageSurface, @SrcRect, S2DLDraw.Surface, @DestRect);
End;

(**
  Draws the Surface onto the Screen. Has been overridden to support multiple
  Different parameter collections. @ This implementation of draw enables the
  display of a section of the image at position X,Y of the S2DLDraw surface. The
  section of the image to display is defined as the pattern number. The number
  of patterns in an image is defined by the PatternHeight and PatternWidth
  properties.

  @param   S2DLDraw as a TS2DLDraw
  @param   X        as an Integer
  @param   Y        as an Integer
  @param   Pattern  as an Integer
**)
Procedure TS2DLImage.Draw(S2DLDraw : TS2DLDraw; X,Y : Integer; Pattern : Integer);
Var
  W,H : Integer;
Begin
  DestRect.x := X;
  DestRect.y := Y;
  SrcRect.W := PatternWidth;
  SrcRect.H := PatternHeight;
  W := ImageSurface.W div PatternWidth;
  H := ImageSurface.H div PatternHeight;
  SrcRect.X := (Pattern Mod W) * PatternWidth;
  SrcRect.Y := (Pattern Div W) * PatternHeight;
  SDL_BlitSurface(ImageSurface, @SrcRect, S2DLDraw.Surface, @DestRect);
End;

Procedure TS2DLImage.Draw(X,Y : Integer);
Begin
  Draw(S2DLDraw,X,Y);
End;

(**
  Draws the Surface onto the Screen. Has been overridden to support multiple
  Different parameter collections. @ This implementation of draw enables the
  display of a section of the image at position X,Y of the base S2DLDraw
  surface. The section of the image to display is defined as the pattern
  number. The number of patterns in an image is defined by the PatternHeight
  and PatternWidth properties.

  @param   X       as an Integer
  @param   Y       as an Integer
  @param   Pattern as an Integer

**)
Procedure TS2DLImage.Draw(X,Y : Integer; Pattern : Integer);
Begin
  Draw(S2DLDraw,X,Y,Pattern);
End;

procedure TS2DLImage.Draw(Surface: PSDL_Surface; X, Y: Integer);
begin
  DestRect.x := X;
  DestRect.y := Y;
  SrcRect.X := 0;
  SrcRect.Y := 0;
  SrcRect.W := ImageSurface.W;
  SrcRect.H := ImageSurface.H;
  SDL_BlitSurface(ImageSurface, @SrcRect, Surface, @DestRect);
end;

procedure TS2DLImage.Draw(Surface: PSDL_Surface; X, Y: Integer; Pattern: Integer
  );
Var
  W,H : Integer;
Begin
  DestRect.x := X;
  DestRect.y := Y;
  SrcRect.W := PatternWidth;
  SrcRect.H := PatternHeight;
  W := ImageSurface.W div PatternWidth;
  H := ImageSurface.H div PatternHeight;
  SrcRect.X := (Pattern Mod W) * PatternWidth;
  SrcRect.Y := (Pattern Div W) * PatternHeight;
  SDL_BlitSurface(ImageSurface, @SrcRect, Surface, @DestRect);
end;

procedure TS2DLImage.DrawTiled(S2DLDraw: TS2DLDraw; X, Y, W, H: Integer);
Var
  OX, I, J : Integer;
begin
  OX := X;
  For I := 0 to H-1 do
    Begin
      X := OX;
      For J := 0 to W-1 do
        Begin
          Draw(S2DLDraw, X,Y);
          X := X + ImageSurface.W;
        End;
      Y := Y + ImageSurface.H;
    End;
end;

procedure TS2DLImage.DrawTiled(S2DLDraw: TS2DLDraw; X, Y, W, H, Pattern: Integer
  );
Var
  OX, I, J : Integer;
begin
  OX := X;
  For I := 0 to H-1 do
    Begin
      X := OX;
      For J := 0 to W-1 do
        Begin
          Draw(S2DLDraw, X,Y, Pattern);
          X := X + PatternWidth;
        End;
      Y := Y + PatternHeight;
    End;
end;

procedure TS2DLImage.DrawTiled(X, Y, W, H: Integer);
begin
  DrawTiled(S2DLDraw, X,Y,W,H);
end;

procedure TS2DLImage.DrawTiled(X, Y, W, H, Pattern: Integer);
begin
  DrawTiled(S2DLDraw, X,Y,W,H,Pattern);
end;

procedure TS2DLImage.DrawTiled(Surface: PSDL_Surface; X, Y, W, H: Integer);
Var
  OX, I, J : Integer;
begin
  OX := X;
  For I := 0 to H-1 do
    Begin
      X := OX;
      For J := 0 to W-1 do
        Begin
          Draw(Surface, X,Y);
          X := X + ImageSurface.W;
        End;
      Y := Y + ImageSurface.H;
    End;
end;

procedure TS2DLImage.DrawTiled(Surface: PSDL_Surface; X, Y, W, H,
  Pattern: Integer);
Var
  OX, I, J : Integer;
begin
  OX := X;
  For I := 0 to H-1 do
    Begin
      X := OX;
      For J := 0 to W-1 do
        Begin
          Draw(Surface, X,Y, Pattern);
          X := X + PatternWidth;
        End;
      Y := Y + PatternHeight;
    End;
end;

Procedure TS2DLImage.TransparentColor(R,G,B : Byte);
Begin
  If ImageSurface <> Nil then
    SDL_SetColorKey(ImageSurface, SDL_SRCCOLORKEY or SDL_RLEACCEL or SDL_HWACCEL,
      SDL_MapRGB(ImageSurface.format, R,G,B));
End;

Type
  (**
    TPixelArray is Used in CopyBitmap to enable a Delphi TBitmap to be copied
    pixel by pixel onto the surface of an S2DLImage.
  **)
  TPixelArray = Array[0..64000] of UInt8;
  (**
    TPixelFormat is Used in CopyBitmap to enable a Delphi TBitmap to be copied
    pixel by pixel onto the surface of an S2DLImage.
  **)
  TPixelFormat = Array[0..2] of UInt8;

procedure TS2DLImage.SetPixel(X, Y: Integer; Color: UInt32);
var
  bits : Pointer;
  bpp, r, g, b : Uint8;
begin
  if ( SDL_LockSurface( ImageSurface ) < 0 ) then
    exit;
  bpp := ImageSurface.format.BytesPerPixel;
  bits := @(TPixelArray(ImageSurface.Pixels^)[(Y * ImageSurface.pitch) + (X * bpp)]);
  // Set the pixel
  case bpp of
      1:
      begin
          PUint8( bits )^ := Uint8(Color);
      end;
      2:
      begin
          PUint16(bits)^ := Uint16(Color);
      end;
      3:
      begin // Format/endian independent
        r := (Color shr ImageSurface.format.Rshift ) and $FF;
        g := (Color shr ImageSurface.format.Gshift ) and $FF;
        b := (Color shr ImageSurface.format.Bshift ) and $FF;
        TPixelFormat(bits^)[(ImageSurface.format.Rshift div 8)] := r;
        TPixelFormat(bits^)[(ImageSurface.format.Gshift div 8)] := g;
        TPixelFormat(bits^)[(ImageSurface.format.Bshift div 8)] := b;
      end;
      4:
      begin
          PUInt32(bits)^ := Uint32( Color );
      end;
  end;
  // Update the display
  SDL_UnlockSurface(ImageSurface);
  SDL_UpdateRect(ImageSurface, X, Y, 1, 1 );
end;

(**
  Enables the drawing of a windows TBitmap object onto the Image Surface. This enables
  the drawing and manipulation of images using the standard windows canvas
  encapulated in a TBitmap to be done and then drawn onto the Images surface.

  @param   X                as an Integer
  @param   Y                as an Integer
  @param   BMP              as a TBitmap
  @param   TransparentColor as an UInt32
**)
procedure TS2DLImage.CopyBitmap(X, Y: Integer; BMP: TBitmap;
  TransparentColor: UInt32);
var
  bits : Pointer;
  I,J, bpp, r, g, b : Uint8;
  Color : UInt32;
begin
  if ( SDL_LockSurface( ImageSurface ) < 0 ) then
    exit;
        bpp := ImageSurface.format.BytesPerPixel;
  For I := 0 to BMP.Width-1 do
    For J := 0 to BMP.Height-1 do
      Begin
        Color := BMP.Canvas.Pixels[I,J];
        If Color <> TransparentColor then
          Begin
            bits := @(TPixelArray(ImageSurface.Pixels^)[((Y+J) * ImageSurface.pitch) + ((X+I) * bpp)]);
            // Set the pixel
            case bpp of
              1:
              begin
                  PUint8( bits )^ := Uint8(Color);
              end;
              2:
              begin
                  PUint16(bits)^ := Uint16(Color);
              end;
              3:
              begin // Format/endian independent
                r := (Color shr ImageSurface.format.Rshift ) and $FF;
                g := (Color shr ImageSurface.format.Gshift ) and $FF;
                b := (Color shr ImageSurface.format.Bshift ) and $FF;
                TPixelFormat(bits^)[(ImageSurface.format.Rshift div 8)] := r;
                TPixelFormat(bits^)[(ImageSurface.format.Gshift div 8)] := g;
                TPixelFormat(bits^)[(ImageSurface.format.Bshift div 8)] := b;
              end;
              4:
              begin
                  PUInt32(bits)^ := Uint32( Color );
              end;
            end;
          End;
  End;
  // Update the display
  SDL_UnlockSurface(ImageSurface);
  SDL_UpdateRect(ImageSurface, X, Y, BMP.Width, BMP.Height);
end;

Initialization
  Images := TS2DLImageList.Create;

Finalization
  Images.Free;

End.
