(** S2DLDraws
@Name: Simple 2D Library - True Type Text Writer
  This file is part of the Free Pascal Simple 2D Libraries
@Version: 1.04
@Note: Do not initialize yourself !!! - Automatically done !!!
@Change_History: ver 1.04 - Check CFG File for Default Font details
                          - Supports SFont - Imaged Fonts
@Copyright: (c) 2004-2005 by William Cairns (CairnsGames) and
                         the Delphi Game Development (DGDev) Site

@License: This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
       (I got this from the JEDI-SDL files and it sounds Cool!)
@
By using this work you have immediatly agreed with the above.
@
 The work can be used for any purpose. You may sell it, use it in a game
 delete it or change it as you wish with no reward to the the original
 developers.
@
 We do however ask that if this work is used that some mention of it be made
 on the PGD web site.
@
 JEDI-SDL ver 1.0 needs changes to the InitFont and PutLine functions.

@Usage:

@Author: W.Cairns **)
unit S2DLText;

{$ifdef FPC}
{$mode delphi}
{$endif}

interface

uses
  Classes, SysUtils, S2DLLogger, sdl, S2DLDraws, sdl_ttf, S2DLImage,
  SFont, SDL_Image;
  
Type
  (**
    TS2DLFont manages the drawing and displaying of text. It can handle both
    ttf (true type fonts) and fonts stored in the standard JEDI-SDL format.
    (look on the net for SFont Builder for more info.
    @
    TTF fonts look different in SDL than in Windows so some testing to work out
    which font is acceptable is required.
    @
    Image Fonts are more difficult to make but make a better user experience.
    They have lower performance impact and can be custom made to suit the game.
  **)
  TS2DLFont = Class;
  (**
    TS2DLFontList manages a set of game fonts.
  **)
  TS2DLFontList = Class
    Private
      FItems : TList;
      Function GetFont(Index : Integer) : TS2DLFont;
      Function GetFontByName(Name : String) : TS2DLFont;
    Public
      Property Items[Index : Integer] : TS2DLFont read GetFont; default;
      Property ItemByName[Name : String] : TS2DLFont read GetFontByName;
      Constructor Create;
      Destructor Destroy; Override;
      Function Find(Name : String) : TS2DLFont;
      Procedure Add(Item : TS2DLFont);
      Function Count : Integer;
      Function Delete(Item : Integer) : Boolean;
      Procedure LoadFont(FontName : String; FontSize : Integer = 24; Bold : Boolean = False);
  End;
  TS2DLFont = Class
    Private
      FBold : Boolean;
      InternalChange : Boolean;
      ExternalChange : Boolean;
      FontSurface : PSDL_Surface;
      FFontColor: TSDL_Color;
      FFontName : PChar;
      FFontSize : Integer;
      FFont : PTTF_Font;
      FText : String;
      FFontType: Integer;
      procedure SetBold(const AValue : Boolean);
      procedure SetFontColor(const AValue: TSDL_Color);
      procedure SetFontName(const AValue: PChar);
      procedure SetFontSize(const AValue: Integer);
      Procedure LoadFont;
      procedure SetFontType(const Value: Integer);
    Public
      Property FontName : PChar read FFontName write SetFontName;
      Property FontSize : Integer read FFontSize write SetFontSize;
      Property FontColor : TSDL_Color read FFontColor write SetFontColor;
      Property Bold : Boolean read FBold write SetBold;
      Property FontType : Integer read FFontType write SetFontType; // 0 = TTF, 1 = Image
      Constructor Create;
      Procedure RenderFont(Surface: PSDL_Surface; X,Y : Integer; Text : String); Overload;
      Procedure RenderFont(X,Y : Integer; Text : String); Overload;
      Procedure Write(Surface: PSDL_Surface; X,Y : Integer; Text : String); Overload;
      Procedure Write(X,Y : Integer; Text : String); Overload;
      Function StaticFont(Text : String) : TS2DLImage;
  End;

Var
  Fonts : TS2DLFontList;

implementation

{ TS2DLFontList }

(**
  This is the constructor method for the TS2DLFontList class.
  Allocates memory to store lists. (Using a list structure).
**)
Constructor TS2DLFontList.Create;
Begin
  FItems := TList.Create;
End;

(**
  This is the destructor method for the TS2DLFontList class.
**)
Destructor TS2DLFontList.Destroy;
Var
  I : Integer;
Begin
  // Free memory used for each image
  For I := FItems.Count -1 downto 0 do
    Begin
      TS2DLFont(FItems[I]).Free;
      FItems.Delete(I);
    End;
  FItems.Free;
End;

(**
  Find the Font identified by Index.
  @see Find
  @param   Index as an Integer
  @return  a TS2DLFont
**)
Function TS2DLFontList.GetFont(Index : Integer) : TS2DLFont;
Begin
  Result := TS2DLFont(FItems[Index]);
End;

(**
  This is a getter method for the FontByName property.
  @param   Name as a String
  @return  a TS2DLFont
**)
Function TS2DLFontList.GetFontByName(Name : String) : TS2DLFont;
Var
  I : Integer;
Begin
  For I := FItems.Count downto 0 do
    Begin
      If TS2DLFont(FItems[I]).FontName = Name then
        Begin
          Result := TS2DLFont(FItems[I]);
          Exit;
        End;
    End;
End;

(**
  Find the Font identified by Name. Return the Font object not an index value.
  This allows the use of Font.Find(FontName).<method>

  @param   Name as a String
  @return  a TS2DLFont
**)
Function TS2DLFontList.Find(Name : String) : TS2DLFont;
Begin
  Result := GetFontByName(Name);
End;

(**
  Add a new font to the Font list.
  @param   Item as a TS2DLFont
**)
Procedure TS2DLFontList.Add(Item : TS2DLFont);
Begin
  FItems.Add(Item);
End;

(**
  Returns the number of fonts available in the list.
  @return  an Integer
**)
Function TS2DLFontList.Count : Integer;
Begin
  Result := FItems.Count;
End;

(**
  Remove a font from the list. This also frees the memory for the font that
  was stored in the list.
  @param   Item as an Integer
  @return  a Boolean
**)
Function TS2DLFontList.Delete(Item : Integer) : Boolean;
Var
  S2DLFont : TS2DLFont;
Begin
  S2DLFont := TS2DLFont(FItems[Item]);
  FItems.Delete(Item);
  S2DLFont.Free;
  Result := True;
End;

(**
  Creates a new TS2DLFont object and loads it into the Font List.
  Then loads the Font specified by FontName into the newly created font.

  @param   FontName as a String
  @param   FontSize as an Integer
  @param   Bold     as a Boolean
**)
procedure TS2DLFontList.LoadFont(FontName: String; FontSize: Integer;
  Bold: Boolean);
Var
  Font : TS2DLFont;
begin
  Font := TS2DLFont.Create;
  Font.FontName := PChar(FontName);
  Font.Bold := Bold;
  Font.FontSize := FontSize;
  Font.LoadFont;
  Fonts.Add(Font);
end;

{ TS2DLFont }

(**
  This is the constructor method for the TS2DLFont class. It automatically loads
  the font specified in the <APPNAME>.cfg file. If no font is defined it will
  load arial.ttf from windows.
**)
constructor TS2DLFont.Create;
Var
  L : TStringList;
  FN : String;
  FS : Integer;
begin
  FN := 'c:\windows\fonts\Arial.ttf';
  FS := 24;
  FontColor := MakeSDLColor(255,0,0);
  If FileExists(Log.ApplicationPath + Copy(Log.ApplicationName, 0, Length(Log.ApplicationName ){$IFDEF WIN32} - 4{$ENDIF} ) + '.cfg') then
    begin
      L := TStringList.Create;
      L.LoadFromFile(Log.ApplicationPath + Copy(Log.ApplicationName, 0, Length(Log.ApplicationName ){$IFDEF WIN32} - 4{$ENDIF} ) + '.cfg');
      FN := L.Values['DefaultFontSize'];
      If FN = '' then FN := '24';
      FS := StrToInt(FN);
      FN := L.Values['DefaultFontName'];
      If FN = '' then
        FN := 'c:\windows\fonts\Arial.ttf';
      L.Free;
    End;
  FontName := PChar(FN);
  FontSize := FS;
  FontColor := MakeSDLColor(255,0,0);
  FText := '';
  InternalChange := True;
  ExternalChange := True;
end;

procedure TS2DLFont.SetFontName(const AValue: PChar);
begin
  if FFontName = AValue then
    exit;
  FFontName := AValue;
  InternalChange := True;
end;

(**
(**
  This is a setter method for the Color property.
  @see MakeSDLColor in S2DLDraws
  @param   AValue as a TSDL_Color constant
**)
procedure TS2DLFont.SetFontColor(const AValue: TSDL_Color);
begin
  If FontType = 1 then
    Exit;
  FFontColor := AValue;
  InternalChange := True;
end;

(**
  This is a setter method for the Bold property.
  When Bold is set the font will print text in bold.
  @note With SDL_ttf this typically means only the font outline is displayed.
  @param   AValue as a Boolean constant
**)
procedure TS2DLFont.SetBold(const AValue : Boolean);
begin
  If FontType = 1 then
    Exit;
  if FBold = AValue
    then exit;
  FBold := AValue;
  ExternalChange := True;
end;

procedure TS2DLFont.SetFontSize(const AValue: Integer);
begin
  If FontType = 1 then
    Exit;
  if FFontSize = AValue then
    exit;
  FFontSize := AValue;
  InternalChange := True;
end;

(**
  Loads a specific font into the object. If FontName contains TTF it is loaded
  as a true type font. All other files are expected to be in SFont format
  images.
**)
procedure TS2DLFont.LoadFont;
begin
  Log.LogStatus('Font name: '+FontName,'TS2DLFont.LoadFont');
  Log.LogStatus('Font size: '+IntToStr(FontSize),'TS2DLFont.LoadFont');
  If Pos('TTF',UpperCase(FontName)) > 0 then
    Begin
      FFont := TTF_OpenFont(FontName, FFontSize);
      If FFont = Nil then
        Begin
          Log.LogError(Format('Couldn''t load font: %s',[SDL_GetError]), 'TS2DLFont.LoadFont');
          Halt(1);
        End;
      InternalChange := False;
      FontType := 0;
    End
  Else
    Begin
      FontSurface := IMG_LOAD(FontName);
      FontType := 1;
    End;
end;

(**
  Writes the text onto the surface, starting at the point (X,Y). Due to the fact
  that a temporary surface has to be created each time the text is rendered S2DL
  caches an image in memory and reuses the cached value unless the text has
  changed. If the same font is used to draw two text strings these strings will
  be recreated each time, using up time.
  @See StaticText
  @param   Surface as a PSDL_Surface
  @param   X       as an Integer
  @param   Y       as an Integer
  @param   Text    as a String
**)
procedure TS2DLFont.RenderFont(Surface: PSDL_Surface; X, Y: Integer; Text: String);
Var
  D : TSDL_Rect;
  RenderStyle : Integer;
  S : PAnsiChar;
begin
  If FontType=1 then
    Begin
      InitFont(FontSurface);
      S := PAnsiChar(Text);
      PutString(Surface,X,Y,S);
      Exit;
    End;
  RenderStyle := TTF_STYLE_NORMAL;
  If InternalChange then
    LoadFont;
  If (FText <> Text) or (ExternalChange) then
    Begin
      SDL_FreeSurface(FontSurface);
      If Bold then
        Begin
          renderstyle := renderstyle or TTF_STYLE_BOLD;
          TTF_SetFontStyle(FFont, renderstyle );
        End;
      FontSurface := TTF_RenderText_Solid(FFont, PChar(Text), FFontColor);
      ExternalChange := False;
    End;
  D.X := X;
  D.Y := Y;
  D.W := FontSurface.W;
  D.H := FontSurface.H;
  if ( SDL_BlitSurface(FontSurface, nil, Surface, @D ) < 0 ) then
    Begin
      Log.LogError(Format('Couldn''t render font: %s',[SDL_GetError]), 'TS2DLFont.RenderFont');
      Halt(1);
    End;
end;

(**
  Writes the text onto the default S2DLDraw surface, starting at the point (X,Y).
  @param   X       as an Integer
  @param   Y       as an Integer
  @param   Text    as a String
**)
procedure TS2DLFont.RenderFont(X, Y: Integer; Text: String);
begin
  RenderFont(S2DLDraw.Surface,X,Y,Text);
end;

(**
  Creates a standard TS2DLImage with the required text. This is used for Headers
  and other text objects that are used regularly but do not change.
  @param   Text as a String
  @return  a TS2DLImage
**)
function TS2DLFont.StaticFont(Text: String) : TS2DLImage;
Var
  D : TSDL_Rect;
  RenderStyle : integer;
  Image : TS2DLImage;
  S : PAnsiChar;
begin
  If FontType=1 then
    Begin
      Image := TS2DLImage.CreateSizedSurface(FontSurface.h * Length(Text),FontSurface.h,16);
      InitFont(FontSurface);
      S := PAnsiChar(Text);
      PutString(Image.ImageSurface,0,0,S);
      Exit;
    End;
  RenderStyle := TTF_STYLE_NORMAL;
  If InternalChange then
    LoadFont;
  If (FText <> Text) or (ExternalChange) then
    Begin
      SDL_FreeSurface(FontSurface);
      If Bold then
        Begin
          renderstyle := renderstyle or TTF_STYLE_BOLD;
          TTF_SetFontStyle(FFont, renderstyle );
        End;
      FontSurface := TTF_RenderText_Shaded(FFont, PChar(Text), FFontColor, MakeSDLColor(0,0,0));
      ExternalChange := False;
    End;
  Log.LogStatus('Draw temp font','Font.StaticText');
  Image := TS2DLImage.CreateSizedSurface(FontSurface.W,FontSurface.H,16);
  Image.TransparentColor(0,0,0);
  Log.LogStatus('Create surface','Font.StaticText');
  D.X := 0;
  D.Y := 0;
  D.W := FontSurface.W-1;
  D.H := FontSurface.H-1;
  SDL_BlitSurface(FontSurface, nil, Image.ImageSurface, @D );
  Log.LogStatus('Draw on surface','Font.StaticText');
  Result := Image;
end;

(**
  Write text onto the surface at position X,Y. Uses RenderFont.
  @param   Surface as a PSDL_Surface
  @param   X    as an Integer
  @param   Y    as an Integer
  @param   Text as a String
**)
procedure TS2DLFont.Write(Surface: PSDL_Surface; X, Y: Integer;
  Text: String);
begin
  RenderFont(Surface,X,Y,Text);
end;

(**
  Write text onto the default S2DLDraw surface at position X,Y. Uses RenderFont.
  @param   X    as an Integer
  @param   Y    as an Integer
  @param   Text as a String
**)
procedure TS2DLFont.Write(X, Y: Integer; Text: String);
begin
  RenderFont(X,Y,Text);
end;

procedure TS2DLFont.SetFontType(const Value: Integer);
begin
  FFontType := Value;
end;

Initialization
  If TTF_Init < 0 then
    Log.LogError(Format('Couldn''t initialize TTF: %s',[SDL_GetError]), 'S2DLText.Initialize');
  Fonts := TS2DLFontList.Create;
    
Finalization
  Log.LogStatus('Closing down fonts', 'S2DLText.Finalization');
  Fonts.Free;
  TTF_Quit;

end.

